#include <stdio.h>
#include <time.h>
#include "cuda.h"

#define WIDTH  16384  // Works with WIDTH<=16384 and doesn't works with WIDTH>16384
#define HEIGHT 10

// Texture
texture<float, 2, cudaReadModeElementType> tex_dev_init;

// Tthe kernel
__global__ void copyFromArrayToGlobal(float* dst){
	unsigned int xIndex = blockIdx.x * blockDim.x + threadIdx.x;
	unsigned int yIndex = blockIdx.y * blockDim.y + threadIdx.y;
	if ( xIndex<WIDTH && yIndex<HEIGHT )
		dst[ yIndex * WIDTH + xIndex] = tex2D(tex_dev_init, xIndex, yIndex);

}

int main(void){

	// For random numbers
	srand(time(NULL));
	
	// Variables
	float     *ptr_host_init;
    cudaArray *array_dev_init;
	float     *ptr_dev_cpy;
	float     *ptr_host_cpy;
	
	// Ramdom initialization of array on host
	ptr_host_init = (float*) malloc( WIDTH * HEIGHT * sizeof(float) );
	for (int i = 0 ; i < WIDTH * HEIGHT ; i++)
		ptr_host_init[i] = (float)rand() / (float)RAND_MAX;
		
	// Declaration of CUDA array and memory copy from host to device
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float>();
    cudaMallocArray( &array_dev_init, &channelDesc, WIDTH, HEIGHT );
    cudaMemcpyToArray( array_dev_init, 0, 0, ptr_host_init, WIDTH * HEIGHT * sizeof(float), cudaMemcpyHostToDevice );
	
	// Bind texture to CUDA array
    tex_dev_init.addressMode[0] = cudaAddressModeClamp;
    tex_dev_init.addressMode[1] = cudaAddressModeClamp;
    tex_dev_init.filterMode     = cudaFilterModePoint;
    tex_dev_init.normalized     = 0;
    cudaBindTextureToArray(tex_dev_init, array_dev_init);

	// Declaration of array on device
	cudaMalloc( (void**) &ptr_dev_cpy, WIDTH * HEIGHT * sizeof(float) );
	
	// Grid and threads
	dim3 grid(WIDTH/16, HEIGHT/16, 1);
	dim3 block(16, 16, 1);
	if (WIDTH  % 16 != 0) grid.x += 1;
	if (HEIGHT % 16 != 0) grid.y += 1;
	
	// Kernel
	copyFromArrayToGlobal<<<grid,block>>>(ptr_dev_cpy);
	
	// Declaration of array on host and memory copy from device to host
	ptr_host_cpy = (float*) malloc( WIDTH * HEIGHT * sizeof(float) );
	cudaMemcpy( ptr_host_cpy, ptr_dev_cpy, WIDTH * HEIGHT * sizeof(float), cudaMemcpyDeviceToHost );
	
	// Verification
	int nb_errors = 0;
	for (int c = 0 ; c < WIDTH ; c++){
		for (int r = 0 ; r < HEIGHT ; r++ ){
			if ( ptr_host_init[r*WIDTH+c] != ptr_host_cpy[r*WIDTH+c] ){
				nb_errors++;
				printf("Error on coordinates (%d,%d) - %f found instead of %f\n",c,r,ptr_host_cpy[r*WIDTH+c],ptr_host_init[r*WIDTH+c]);
			}
		}
	}
	printf("Number of errors : %d\n",nb_errors);
	if (nb_errors)
		printf("FAILED\n");
	else
		printf("PASSED\n");

	// Free memory
	free(ptr_host_cpy);
	cudaFree(ptr_dev_cpy);
	cudaFreeArray(array_dev_init);
	free(ptr_host_init);
}